package game;

import java.util.Random;
import java.util.Vector;

/**
 * Mission.java
 * An instance of a game mission.
 * @author Noda
 */
public final class Mission {
    
    // constants
    public static final int NUM_ROOMS = 10;     // number of rooms in the spacecraft
    public static final int[] ROOM_COORD = {    // coordinate of rooms
        239, 76,        // command room
        185, 258,       // laboratory
        295, 258,       // playing room
        239, 156,       // briefing room
        185, 364,       // computer room
        295, 364,       // rest room
        175, 450,       // kitchen garden
        305, 450,       // waters
        241, 513,       // garden
        241, 591        // engine room
    };
    public static final int[] CAN_MOVE_TO = {   // links between rooms
        1, 3, 2,        // command room doors
        0, 3, 4,        // laboratory doors
        3, 0, 5,        // playing room doors
        0, 1, 2,        // briefing room doors
        1, 5, 6,        // computer room doors
        4, 2, 7,        // rest room doors
        4, 8, 9,        // kitchen doors
        8, 5, 9,        // waters doors
        6, 7, 9,        // garden doors
        6, 7, 8         // engine room doors
    };
    public static final String[] ROOM_AMBIANCE = {
        "/sounds/rooms/command_room.mp3",
        "/sounds/rooms/laboratory.mp3",
        "/sounds/rooms/playing_room.mp3",
        "/sounds/rooms/briefing.mp3",
        "/sounds/rooms/computer_room.mp3",
        "/sounds/rooms/rest_room.mp3",
        "/sounds/rooms/kitchen.mp3",
        "/sounds/rooms/waters.mp3",
        "/sounds/rooms/garden.mp3",
        "/sounds/rooms/engine_room.mp3",
    };
    
    // attributes
    public MissionType type;                    // type of the mission
    public boolean[] destroyed;                 // states of different rooms
    public boolean[] activity;                  // activity (repair or destroy) in rooms
    public int phase;                           // current phase of the mission
    public int turn;                            // current turn of the mission
    public Player[] players;                    // players of the mission for the server (a client use only players[0]), 0 is the player on this phone
    public int[] votes;                         // vote results
    public int eliminate;                       // id of the player to eliminate this turn
    public boolean isServer;                    // this machine is the server
    public int connected;                       // players connected (number of clients)
    public int playersReady;                    // players ready at RDV
    public boolean ready;                       // ready to go to next phase
    public Random rand;                         // random number generator
    public int commander;                       // id of the commander
    public Vector vectReader;
    public Writer writer;
    
    /** creates a new instance of Mission */
    public Mission(boolean isServer, MissionType type, int connected,Vector r,Writer w) {
        vectReader=r;
        writer=w;
        phase = MissionPhase.INIT_PRESENTATION;
        turn = 0;
        eliminate = -1;
        this.type = type;
        this.isServer = isServer;
        if (isServer) this.connected = connected;
        else this.connected=1;
        players = new Player[type.num_players];
        votes = new int[type.num_players+1];
        destroyed = new boolean[NUM_ROOMS];
        activity = new boolean[NUM_ROOMS];
        rand = new Random();
        initSession();
    }
    
    /** return the number of people alive */
    public final int getAlive() {
        int nb = 0;
        for (int i=0; i < type.num_players; i++)
            if (players[i].alive)
                ++nb;
        return nb;
    }
    
    /** return the number of rooms destroyed */
    public final int getDestroyed() {
        int nb = 0;
        for (int i=0; i < NUM_ROOMS; i++)
            if (destroyed[i])
                ++nb;
        return nb;
    }
    
    /** return the number of technicians alive */
    public final int getTechnicians() {
        int nb = 0;
        for (int i=0; i < type.num_players; i++)
            if (players[i].type == PlayerType.TECHNICIAN && players[i].alive)
                ++nb;
        return nb;
    }
    
    /** return the number of wreckers alive */
    public final int getWreckers() {
        int nb = 0;
        for (int i=0; i < type.num_players; i++)
            if (players[i].type == PlayerType.WRECKER && players[i].alive)
                ++nb;
        return nb;
    }
    
    /** return the number of aliens alive */
    public final int getAliens() {
        int nb = 0;
        for (int i=0; i < type.num_players; i++)
            if (players[i].type == PlayerType.ALIEN && players[i].alive)
                ++nb;
        return nb;
    }
    
    /** return the max possible vote id + 1 for a given player id */
    public final int getMaxVote(int p_id) {
        int nb = 1;
        for (int i=1; i<type.num_players; i++)
            if (players[i].alive)
                ++nb;
        
        if (players[p_id].type == PlayerType.COMMANDER)
            return nb;
        else
            return nb-1;
    }
    
    /** return the player id corresponding to the given vote id and current player id */
    public final int getPlayerID(int vote, int curr_player) {
        int p_id = -1;
        if (vote == 0)
            return p_id;
        
        int j=1;
        for (int i=0; i < type.num_players; i++) {
            if ((i != curr_player) && (players[i].type != PlayerType.COMMANDER) && (players[i].alive)) {
                if (j == vote) {
                    p_id = i;
                    break;
                }
                j++;
            }
        }
        return p_id;
    }
    
    /** function to be called at each frame */
    public final void tick() {
        if (playersReady == connected)
            ready = true;
    }
    
    /** return true if the current player won the game */
    public final boolean getWin() {
        boolean win = false;
        int type = players[0].type;
        switch(MissionType.checkWin(getAliens(), getTechnicians(), getWreckers(), getDestroyed(), turn)) {
            case MissionType.GOODTEAM_WINS:
                if (type == PlayerType.COMMANDER || type == PlayerType.TECHNICIAN)
                    win = true;
                break;
            case MissionType.ALIENS_WINS:
                if (type == PlayerType.ALIEN)
                    win = true;
                break;
            case MissionType.WRECKERS_WINS:
                if (type == PlayerType.WRECKER)
                    win = true;
                break;
        }
        return win;
    }
    
    /** advance to next phase of the mission */
    public final void nextPhase() {
        if (isServer) {
            AIPlayAll();
            playersReady = 0;
            ready = false;
        } else {
            playersReady = 0;
            ready = false;
        }
        
        phase++;
        // loop phases if it's not the last turn
        if (phase == MissionPhase.END) {
            if (eliminate != -1) {
                if (isServer)
                    players[eliminate].alive = false;
                else
                    players[eliminate].alive = false;
            }
            phase = MissionPhase.MOVE;
            ++turn;
            // check for win game case
            if (MissionType.checkWin(getAliens(), getTechnicians(), getWreckers(), getDestroyed(), turn) != MissionType.NOT_FINISHED)
                phase = MissionPhase.END;
            // end if it player is alive
            if (!players[0].alive) {
                phase = MissionPhase.END;
            } else {
                for (int i=0; i < NUM_ROOMS; i++)
                    activity[i] = false;
            }
        } else if (phase == MissionPhase.COMMANDER_ACTION) {
            eliminate = -2;
            // count votes of all players
            for (int i=0; i <= type.num_players; i++)
                votes[i] = 0;
            for (int i=0; i < type.num_players; i++)
                if (players[i].alive)
                    ++votes[players[i].vote+1];
            // wait for reception of the commander decision by the clients
            if (!isServer){
                Reader r =(Reader)vectReader.elementAt(0);
                r.wake(3,this);
            }
        } else if (phase == MissionPhase.STATS) {
            // check for win game case
            if (MissionType.checkWin(getAliens(), getTechnicians(), getWreckers(), getDestroyed(), turn) == MissionType.WRECKERS_WINS)
                phase = MissionPhase.END;
        } else if (phase == MissionPhase.WAIT_PLAYERS0){
            // first synchro phase
            if (connected!=0 && isServer){
                for (int i=0;i<vectReader.size();i++) {
                    Reader r =(Reader)vectReader.elementAt(i);
                    r.wake(0,this);
                }
            } else if (!isServer) {
                Reader r =(Reader)vectReader.elementAt(0);
                r.wake(0,this);
            }
        } else if (phase == MissionPhase.WAIT_PLAYERS1){
            // second synchro phase
            if (connected!=0 && isServer){
                for (int i=0;i<vectReader.size();i++) {
                    Reader r =(Reader)vectReader.elementAt(i);
                    r.wake(1,this);
                }
            } else if (!isServer) {
                Reader r =(Reader)vectReader.elementAt(0);
                r.wake(1,this);
            }
        } else if (phase == MissionPhase.WAIT_PLAYERS2) {
            // third synchro phase
            if (connected!=0 && isServer) {
                for (int i=0;i<vectReader.size();i++) {
                    Reader r =(Reader)vectReader.elementAt(i);
                    r.wake(2,this);
                }
            } else if (!isServer) {
                Reader r =(Reader)vectReader.elementAt(0);
                r.wake(2,this);
            }
        } else if (phase == MissionPhase.DECISION) {
            // send commander decision by the server
            if (connected!=0 && isServer){
                for (int i=0;i<vectReader.size();i++) {
                    writer.wake(0,this);
                }
            }
        }
    }
    
    /** play phase for all AI-controlled alive players */
    private final void AIPlayAll() {
        for(int i=0; i < (type.num_players - connected - 1); i++)
            if (players[type.num_players-i-1].alive)
                AIPlay(type.num_players-i-1);
    }
    
    /** return the number of playing who haven't finished their phase */
    public final int getNumWait() {
        return (connected - playersReady);
    }
    
    /** play phase of an AI-controlled player */
    private final void AIPlay(int p_id) {
        try {
            Player p = players[p_id];
            switch (phase) {
                case MissionPhase.MOVE:
                    // move randomly to an adjacent room, or stay in place
                    int moveTo = rand.nextInt(4);
                    if (moveTo != 3)
                        p.inRoom = Mission.CAN_MOVE_TO[p.inRoom*3 + moveTo];
                    break;
                case MissionPhase.ACTION:
                    // do an action corresponding to player's role
                    int actionToDo = MissionType.ACTION_NOACTION;
                    // get the action
                    switch (p.type) {
                        case PlayerType.COMMANDER:
                            break;
                        case PlayerType.TECHNICIAN:
                            if (rand.nextInt(100) < 90)
                                actionToDo = MissionType.getActionCode(p, destroyed[p.inRoom] && !activity[p.inRoom], 1);
                            break;
                        case PlayerType.WRECKER:
                            int base;
                            if (players[0].type == PlayerType.WRECKER && connected == 0)
                                base = 60;
                            else
                                base = 50;
                            if (rand.nextInt(100) < (base + (100-base) * ((float)turn / (float)(type.num_turns-1))))
                                actionToDo = MissionType.getActionCode(p, destroyed[p.inRoom], 1);
                            break;
                        case PlayerType.ALIEN:
                            if (rand.nextInt(100) < (20 + 50 * ((float)turn / (float)(type.num_turns-1))))
                                actionToDo = MissionType.getActionCode(p, destroyed[p.inRoom] && !activity[p.inRoom], 1);
                            break;
                    }
                    // do the action
                    if (actionToDo != MissionType.ACTION_NOACTION) {
                        if (actionToDo == MissionType.ACTION_DESTROY)
                            destroyed[p.inRoom] = true;
                        else if (actionToDo == MissionType.ACTION_REPAIR)
                            destroyed[p.inRoom] = false;
                        activity[p.inRoom] = true;
                        p.actions_left[actionToDo]--;
                    }
                    break;
                case MissionPhase.VOTE:
                    // chance for the good team to vote right is increasing with turn,
                    // for the vilains they vote randomly
                    int vote = 0;
                    switch (p.type) {
                        case PlayerType.COMMANDER:
                        case PlayerType.TECHNICIAN:
                            int base, max;
                            if ((players[0].type == PlayerType.WRECKER || players[0].type == PlayerType.ALIEN) && connected == 0) {
                                base = 25;
                                max = 75;
                            } else {
                                base = 60;
                                max = 90;
                            }
                            if (rand.nextInt(100) < (base + (max-base) * ((float)turn / (float)(type.num_turns-1)))) {
                                do {
                                    do {
                                        vote = getPlayerID(rand.nextInt(getMaxVote(p_id)), p_id);
                                    } while (vote == -1);
                                } while (players[vote].type != PlayerType.WRECKER && players[vote].type != PlayerType.ALIEN);
                            } else {
                                vote = getPlayerID(rand.nextInt(getMaxVote(p_id)), p_id);
                            }
                            break;
                        case PlayerType.WRECKER:
                            vote = getPlayerID(rand.nextInt(getMaxVote(p_id)), p_id);
                            break;
                        case PlayerType.ALIEN:
                            vote = getPlayerID(rand.nextInt(getMaxVote(p_id)), p_id);
                            break;
                    }
                    p.vote = vote;
                    break;
                case MissionPhase.COMMANDER_ACTION:
                    int base;
                    if ((players[0].type == PlayerType.WRECKER || players[0].type == PlayerType.ALIEN) && connected == 0) {
                        base = 20;
                    } else {
                        base = 30;
                    }
                    // only do something if the player is the commander
                    if (p.type == PlayerType.COMMANDER) {
                        // double count its vote
                        ++votes[p.vote+1];
                        // get the player who has most votes
                        int pm_id = 0, max = votes[pm_id];
                        for (int i=1; i < type.num_players+1; i++)
                            if (votes[i] >= max) {
                            pm_id = i;
                            max = votes[i];
                            }
                        // get the confidence % for that player
                        float confidence = (max*100.0f) / (float)getAlive();
                        // vote accordingly to that confidence %
                        if (rand.nextInt(100) < confidence) {
                            // the chance to vote right increase with turns
                            if (rand.nextInt(100) < (base + 30 * ((float)turn / (float)(type.num_turns-1)))) {
                                if (pm_id != 0 && (players[pm_id-1].type == PlayerType.WRECKER || players[pm_id-1].type == PlayerType.ALIEN)) {
                                    eliminate = pm_id-1;
                                } else {
                                    // last chance to vote right, is if the commander vote is right
                                    if (p.vote != -1)
                                        if (players[p.vote].type == PlayerType.WRECKER || players[p.vote].type == PlayerType.ALIEN)
                                            eliminate = p.vote;
                                        else
                                            eliminate = -1;
                                    else
                                        eliminate = -1;
                                }
                            } else {
                                eliminate = pm_id-1;
                            }
                        } else {
                            if (rand.nextInt(100) < 50)
                                // blank vote
                                eliminate = -1;
                            else
                                eliminate = pm_id-1;
                        }
                    }
                    break;
            }
        } catch(Exception e) { 
            System.err.println(e.getMessage()); 
        }
        
    }
    
    /** set all players to a ready state */
    public synchronized boolean setPlayersReady(boolean last){
        if (last) {
            playersReady++;
            return true;
        } else if (playersReady==connected-1)
            return true;
        else {
            playersReady++;
            return false;
        }
    }
    
    /** write all rooms information */
    public void writeRoom(){
        byte[] b = new byte[7];
        for (int i=1; i<7; i++) {
            b[i] = (byte)players[i-1].inRoom;
            System.out.println("Room " + (int)b[i]);
        }
        writer.write(b);
    }
    
    /** write activity information */
    public void writeActivity(){
        byte[] b = new byte[21];
        int i = 1;
        for (int j=0; j<10; j++) {
            if (activity[j])
                b[i]=(byte)1;
            else
                b[i]=(byte)0;
            i++;
            if (destroyed[j])
                b[i]=(byte)1;
            else
                b[i]=(byte)0;
            i++;
        }
        writer.write(b);
    }
    
    /** write all vote information */
    public void writeVote(){
        byte[] b=new byte[1];
        b[0]=0;
        writer.write(b);
    }
    
    /** init the game session */
    private final void initSession() {
        if(isServer) {
            // **** init server ****
            int avatar_id;
            int[] name_id = new int[type.num_players];
            int role;
            int num_commander = 1;
            int num_wreckers = type.num_wreckers;
            int num_aliens = type.num_aliens;
            int num_technicians = type.num_technicians;
            int[] room = new int[NUM_ROOMS];
            String name;
            boolean ok;
            
            // init players
            for(int i=0; i<type.num_players; i++) {
                // choose a random, unused avatar
                do {
                    ok = false;
                    avatar_id = rand.nextInt(Player.NUM_AVATAR);
                    for(int j=0; j<i; j++)
                        if(avatar_id == players[j].avatar && i != j)
                            ok = true;
                } while(ok);
                // choose a random, unused name
                do {
                    ok = false;
                    name_id[i] = rand.nextInt(Player.NUM_NAMES);
                    for(int j=0; j<i; j++)
                        if(name_id[i] == name_id[j] && i != j)
                            ok = true;
                } while(ok);
                // choose a random, unused role
                do {
                    ok = false;
                    if (i == 0 && connected > 0)
                        role=PlayerType.COMMANDER;
                    else
                        role = rand.nextInt(PlayerType.NUM_TYPES);
                    switch(role) {
                        case PlayerType.COMMANDER: if(num_commander == 0) ok = true; else { num_commander--; commander = i; }  break;
                        case PlayerType.ALIEN: if(num_aliens == 0) ok = true; else num_aliens--;  break;
                        case PlayerType.TECHNICIAN: if(num_technicians == 0) ok = true; else num_technicians--;  break;
                        case PlayerType.WRECKER: if(num_wreckers == 0) ok = true; else num_wreckers--;  break;
                    }
                } while(ok);
                // choose a random, unused role
                do {
                    ok = false;
                    room[i] = rand.nextInt(NUM_ROOMS);
                    for(int j=0; j<i; j++)
                        if(room[i] == room[j] && i != j)
                            ok = true;
                } while(ok);
                // name of the player
                if(avatar_id < Player.NUM_AVATAR/2)
                    name = Player.NAMES_M[name_id[i]];
                else
                    name = Player.NAMES_F[name_id[i]];
                // create the player
                players[i] = new Player(role, avatar_id, name, room[i], i);
                switch(role) {
                    case PlayerType.COMMANDER:
                        break;
                    case PlayerType.ALIEN:
                        players[i].actions_left[MissionType.ACTION_DESTROY] = 1;
                        players[i].actions_left[MissionType.ACTION_REPAIR] = 1;
                        break;
                    case PlayerType.TECHNICIAN:
                        players[i].actions_left[MissionType.ACTION_REPAIR] = 2;
                        break;
                    case PlayerType.WRECKER:
                        players[i].actions_left[MissionType.ACTION_DESTROY] = -1;
                        break;
                }
                
            }
            // send initialization of the game to every connected client
            if (connected!=0){
                byte[] b = new byte[100];
                int j=1;
                for (int i=0; i<type.num_players; i++) {
                    b[j]=(byte)players[i].type;
                    System.out.print("role "+players[i].type);
                    j++;
                    b[j]=(byte)players[i].avatar;
                    System.out.print("avatar "+players[i].avatar);
                    j++;
                    b[j]=(byte)players[i].inRoom;
                    System.out.print("room "+players[i].inRoom);
                    j++;
                    b[j]=(byte)players[i].server_id;
                    System.out.print("server_id "+players[i].server_id);
                    j++;
                    byte[] chars= players[i].name.getBytes();
                    System.out.println(players[i].name);
                    for(int k=0;k<chars.length;k++){
                        b[j]=chars[k];
                        j++;
                    }
                }
                writer.write(b);
            }
            playersReady = 0;
            ready = false;
            
        } else {
            // recept initialisation of the game by a client
            Reader r=(Reader)vectReader.elementAt(0);
            byte[] b = new byte[100];
            b=r.readBytes(100);
            int l=0;
            int j=1;
            int k;
            int client_id=(int)b[l];l++;
            // init players
            for(int i=0; i<type.num_players; i++) {
                int n=0;
                int role=(int)b[l++];
                int avatar_id=(int)b[l++];
                int room=(int)b[l++];
                int server_id=(int)b[l++];
                String name=null;
                n=0;
                char[] nom=new char[20];
                nom[0]=(char)b[l];
                while (nom[n]!='.') {
                    l++;
                    n++;
                    nom[n]=(char)b[l];
                }
                name=new String(nom,0,n+1);
                l++;
                if (i==client_id){
                    // create the player
                    k=0;
                    players[k] = new Player(role, avatar_id, name, room, server_id);
                } else {
                    k=j;
                    players[k] = new Player(role, avatar_id, name, room, server_id);
                    j++;
                }
                switch(role) {
                    case PlayerType.COMMANDER:
                        commander=k;
                        break;
                    case PlayerType.ALIEN:
                        players[k].actions_left[MissionType.ACTION_DESTROY] = 1;
                        players[k].actions_left[MissionType.ACTION_REPAIR] = 1;
                        break;
                    case PlayerType.TECHNICIAN:
                        players[k].actions_left[MissionType.ACTION_REPAIR] = 2;
                        break;
                    case PlayerType.WRECKER:
                        players[k].actions_left[MissionType.ACTION_DESTROY] = -1;
                        break;}
            }
            playersReady = 0;
            ready = false;
            for(int i=0; i<type.num_players; i++) {
                System.out.print("role "+players[i].type);
                System.out.print("avatar "+players[i].avatar);
                System.out.print("room "+players[i].inRoom);
                System.out.print("server_id "+players[i].server_id);
                System.out.println(players[i].name);
            }
        }
    }
}